/**
 * @file genericdll.c
 * Template for creating plugins
 */

#include "../dllkit/dllkit.h"
#include "../usrmsgs.h"
#include <stdio.h>
#include <tchar.h>

/** @brief Handle to instance of this module */
HANDLE hThisModule;
/** @brief Contains list of messages to be forwarded to this plugin */
MSGLIST Msgs;
/** @brief Local copy of a structure passed to this plugin containging
 * various info on the main emu6502 window and application */
INITDLLSTRUCT info;

/** @brief Entry point of this dll */
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	// Save copy of handle to this module
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		// Create list of messages to be forwarded to this plugin
		CreateMsgList(&Msgs, 2);
		break;
	case DLL_PROCESS_DETACH:
		// Free this list
		FreeMsgList(&Msgs);
		break;
	}	
    return TRUE;
}

/**
 * @see InitDll in the dllkit.h
 */
DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	// Fill the msglist with messages we want handle
	AddMessage(Msgs, CPURUNNING);
	AddMessage(Msgs, CPUSTOPPED);

	// Copy informations passed to this dll
	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	info.optional = ids->optional;
	info.hTb = ids->hTb;


	// Fill the return idi structure
	// This reserves IO memory area IOLO + 21 to IOLO + 30 for this plugin
	// After accesing this area, IOFunc will be called
	idi->iolo = IOLO + 21;
	idi->iohi = IOLO + 30;
	idi->msglist = Msgs;
}

/**
 * @see OnMessage in the dllkit.h
 */
DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	switch (mds->message) {
	case CPURUNNING:
		Running = TRUE;
		break;
	case CPUSTOPPED:
		Running = FALSE;
		break;
	}

}

/** Maybe useful macro for IOFunc */
#define IOSETGET(GVAR, SVAR) \
	if (mode == IOREAD) {	\
		*data = GVAR;	\
	}	\
	else {	\
		SVAR = *data;	\
	}


/**
 * @see InitDll in the dllkit.h
 */
DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, iomode mode)
{
	switch (adr) {
	case IOLO + 21:
		if (mode == IOREAD) {
			// set *data according to local state
		}
		else {
			// set local variables / perform some actions accoring to *data
		}
		break;
	}
}

