/**
 * @file mem.c
 * Memory management is done in this file. It controls shared acces
 * to the memory structure via the CRITICAL_SECTION synchronization
 * primitive.
 * In trouble you can also uncomment the define DEBUG directive.
 * It will create an out.txt file in which all accesses into memory
 * will be written (except 0xA000 - 0xD000, but you can comment this)
 */

#include "mem.h"
#include <memory.h>
#include <windows.h>

#include <stdio.h>

//#define DEBUG

#ifdef DEBUG
#define POCET 1000000
typedef struct { BOOL read; UINT adr; unsigned char data; } item;
item DArr[POCET];
FILE* fil;
int volatile uk = 0;
#endif

unsigned char memory[MEMSIZE + 4];	
pIOFunc memio[IOHI - IOLO + 1];

CRITICAL_SECTION CSMemory;

unsigned char memgetb(UINT adr)
{
	unsigned char bget;
	EnterCriticalSection(&CSMemory);
#ifdef DEBUG
	if (uk>POCET) {
		int i;
		fil = fopen("out.txt", "a");
		for (i = 0; i<= POCET; i++) {
			if (DArr[i].read == TRUE) {
				fprintf(fil, "cti(%04x) = %02x\n", DArr[i].adr, DArr[i].data);
			} else {
				fprintf(fil, "nastav(%04x)<---%02x\n", DArr[i].adr, DArr[i].data);
			}
			fflush(fil);
		}
		uk = 0;
		fflush(fil);
		fclose(fil);
	}
	if (adr < 0xA000 || adr >= 0xD000){// fprintf(fil, "cti(%x)=", adr);
	DArr[uk].adr = adr;
	DArr[uk].read = TRUE;
	}
#endif
	if (adr < IOLO || adr > IOHI || memio[adr - IOLO] == NULL) {
		bget = memory[adr];
	}
	else {
		(memio[adr - IOLO])(adr, &bget, IOREAD);
	}
#ifdef DEBUG
	if (adr < 0xA000 || adr >= 0xD000){// fprintf(fil, "%x\n", bget);
	DArr[uk].data = bget;
	uk++;
	}
#endif
	LeaveCriticalSection(&CSMemory);
	return bget;
}

void memsetb(UINT adr, unsigned char bset)
{
	EnterCriticalSection(&CSMemory);
#ifdef DEBUG
	if (adr < 0xA000 || adr >= 0xD000){// fprintf(fil, "nastav(%x)<--", adr);
	DArr[uk].adr = adr;
	DArr[uk].read = FALSE;
	}
#endif
	if (adr < IOLO || adr > IOHI || memio[adr - IOLO] == NULL) {
		memory[adr] = bset;
	}
	else {
		(memio[adr - IOLO])(adr, &bset, IOWRITE);
	}
#ifdef DEBUG
	if (adr < 0xA000 || adr >= 0xD000){// fprintf(fil, "%x\n", bset);
	DArr[uk].data = bset;
	uk++;
	}
#endif
	LeaveCriticalSection(&CSMemory);
}

void InitMem()
{
	UINT i;
#ifdef DEBUG
	//fil = fopen("out.txt", "w");
#endif
	for (i = 0; i <= IOHI - IOLO; i++) {
		memio[i] = NULL;
	}
	InitializeCriticalSection(&CSMemory);
}

void SetIOFunc(pIOFunc piof, UINT loadr, UINT hiadr)
{
	UINT i;
	for (i = loadr - IOLO; i <= hiadr - IOLO; i++) {
		memio[i] = piof;
	}
}

void SetMemory(char *buf)
{
	memcpy(memory, buf, MEMSIZE);
}

#ifdef DEBUG
void printmsg(char* msg)
{
	fprintf(fil, "%s", msg);
}
#endif