/**
 * @file mem.h
 * .h file to the @a mem.c
 */

#ifndef MEMORY_H
#define MEMORY_H

#define TRUE 1
#define FALSE 0

#include "dlluser.h"

typedef unsigned int UINT;

/** @brief Size of memory */
#define MEMSIZE 65536
/** @brief No. of bits of pagesize. Should correspond with @a MEMSIZE and @a PAGECOUNT. */
#define PAGESIZE 8
/** @brief No. of pages. Should correspond with @a MEMSIZE and @a PAGECOUNT. */
#define PAGECOUNT 256

/** @brief Lower boundary of memory serving for device maping */
#define IOLO 0xD000
/** @brief Higher boundary of memory serving for device maping */
#define IOHI 0xD7FF

/** @brief Variable containing accessible memory +4 as pillow */
extern unsigned char memory[];

/** @brief A synchronization primitive used for shared acces to memory */
extern CRITICAL_SECTION CSMemory;

/** 
 * @brief IO part of memory 
 *
 * Containins pointers to io functions of devices mapped into 
 * the io part of memory.
 * @see iofunc
 */
extern pIOFunc memio[IOHI - IOLO + 1];

/** 
 * @brief memory reading function
 *
 * Calls iofunc when neccesary
 * @see memsetb
 */
unsigned char memgetb(UINT adr);

/** 
 * @brief memory writing function
 *
 * Calls iofunc when neccesary
 * @see memgetb
 */
void memsetb(UINT adr, unsigned char byte);

/** @brief Fills @a memory with given data */
void SetMemory(char *buf);

/** @brief Zeroes @a memio and initializes CSMemory*/
void InitMem();

/** 
 * @brief Binds given io function to range of adresses
 *
 * @param piof An io function which should service reading and writing 
 * to these adresses
 * @param loadr Lower end of the range demanded
 * @param hiadr Higher end of the range demanded
 */
void SetIOFunc(pIOFunc piof, UINT loadr, UINT hiadr);


#endif //#ifndef MEMORY_H
