/**
 * @file perifdll.c
 * Handles events from misc periferies (at now keyboard only) and also
 * contains random number generator
 */

#include "../dllkit/dllkit.h"
#include "../usrmsgs.h"
#include <tchar.h>
#include <stdlib.h>	//Random

#define KBSTAT IOLO + 21
#define SHIFTSTAT IOLO + 22
#define RANDOM IOLO + 23

#define ADR(X, Y) (((X) << 8) + (Y))

/** 
 * @brief Translation table 
 * 
 * Translates Windows Virtual Keys to ATASCII codes
 * doesn't handles all keys, to acces eg. small caps, you have to
 * do it in the emulator's operating system (which isn't done by me :o)
 */
unsigned char KbMap[256] = { 
//	1  , 2  , 3  , 4  , 5  , 6  , 7  , 8  , 9  , 10
	255, 255, 255, 255, 255, 255, 255,  30, 125, 255,	// 10
	255, 255,   0, 255, 255, 255, 255, 255, 255,   1,	// 20
	255, 255,   2,   3,	  4,   5,  27, 255, 255, 255,	// 30
	255,  32, 255, 255, 255, 255,  30,  28,  31,  29,	// 40
	255, 255, 255, 255, 127, 126, 255,  48,  49,  50,	// 50
	 51,  52,  53,  54,  55,  56,  57, 255, 255, 255,	// 60
	255, 255, 255, 255,  65,  66,  67,  68,  69,  70,	// 70
	 71,  72,  73,  74,  75,  76,  77,  78,  79,  80,	// 80
	 81,  82,  83,  84,  85,  86,  87,  88,  89,  90,	// 90
	255, 255, 255, 255, 255,  48,  49,  50,  51,  52,	// 100
	 53,  54,  55,  56,  57,  42,  43, 255,  45, 255,	// 110
	 47, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 120
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 130
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 140
	255, 255, 255, 124, 255, 255, 255, 255, 255, 255,	// 150
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 160
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 170
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 180
	255, 255, 255, 255, 255,  59,  61,  44,  45,  46,	// 190
	 47, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 200
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 210
	255, 255, 255, 255, 255, 255, 255, 255,  91,  92,	// 220
	 93,  39, 255, 255, 255, 255, 255, 255, 255, 255,	// 230
	255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// 240
	255, 255, 255, 255, 255, 255, 255
};

#define SHIFTFLAG 0x01
#define CTRLFLAG 0x02


volatile unsigned char Key = 0;
volatile unsigned char KbState = 0;


HANDLE hThisModule;
MSGLIST Msgs;
INITDLLSTRUCT info;

BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		CreateMsgList(&Msgs, 2);
		break;
	case DLL_PROCESS_DETACH:
		FreeMsgList(Msgs);
		break;
	}	
    return TRUE;
}

DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	// Add Message list of messages to be catched
	AddMessage(Msgs, WM_KEYDOWN);
	AddMessage(Msgs, WM_KEYUP);

	// Copy informations passed to this dll
	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	info.optional = ids->optional;


	Key = 0;
	 
	// Initialize pseudo-random number generator 
	srand((UINT) GetTickCount());

	// Fill the return idi structure
	idi->iolo = IOLO + 21;
	idi->iohi = IOLO + 30;
	idi->msglist = Msgs;
}


DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	switch (mds->message) {
	case WM_KEYDOWN:
		switch (mds->wParam) {
		case VK_SHIFT:
			KbState |= SHIFTFLAG;
			break;
		case VK_CONTROL:
			KbState |= CTRLFLAG;
			break;
		default:
			Key = (char) mds->wParam & 0xFF;
			if (KbMap[Key - 1] != 255) {
				(*info.memsetb)(IRQID, 1);
				SendMessage(info.hWnd, IRQMSG, 0, 0);
			}
			break;
		}
		break;
	case WM_KEYUP:
		switch (mds->wParam) {
		case VK_SHIFT:
			KbState &= ~SHIFTFLAG;
			break;
		case VK_CONTROL:
			KbState &= ~CTRLFLAG;
			break;
		default:
			if (Key != 0 && KbMap[Key - 1] != 255) {
				(*info.memsetb)(IRQID, 2);
				SendMessage(info.hWnd, IRQMSG, 0, 0);
			}
			Key = 0;
			break;
		}
		break;
	}

}

#define IOSETGET(GVAR, SVAR) \
	if (mode == IOREAD) {	\
		*data = GVAR;	\
	}	\
	else {	\
		SVAR = *data;	\
	}

DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, 
										  iomode mode)
{
	switch (adr) {
	case KBSTAT:
		if (mode == IOREAD && Key != 0) {
			*data = KbMap[Key - 1];
		}
		else if (mode == IOREAD) {
			*data = 255;
		}
		//*data = 75;
		break;
	case SHIFTSTAT:
		if (mode == IOREAD) {
			*data = KbState;
		}
		break;
	case RANDOM:
		if (mode == IOREAD) {
			*data = (unsigned char) (rand() % 0xFF);
		}
		break;
	}
}

