/**
 * @file usrmsgs.h
 * Common file for all plugins. 
 * Here are defined all user messages used by the plugins.
 * Also serves as a list of reserved message numbers
 */

#ifndef USRMSGS_H
#define USRMSGS_H

/** @brief Message sent to the msg. queue when NMI interrput occurs */
#define NMIMSG WM_USER + 0
/** @brief Message sent to the msg. queue when IRQ interrput occurs */
#define IRQMSG WM_USER + 1
/** @brief Message sent to the msg. queue when RESET interrput occurs */
#define RESETMSG WM_USER + 2
/** @brief lParam of the message is desired frequency in Hz */
#define SETFREQ WM_USER + 3
/** @brief lParam retrieves adress of Regs struct, where registers are stored. lParam should point to some UINT */
#define GETREGS WM_USER + 4
/** @brief Returns TRUE if running otherwise FALSE */
#define ISRUNNING WM_USER + 5
/** @brief Executes one instruction. lParam should point to some UINT. lParam retrieves no. of cycles the instruction has taken. */
#define ISTEP WM_USER + 6
/** @brief A debug message, execute this interrupt immediately */
#define IRQD WM_USER + 7
/** @brief A debug message, execute this interrupt immediately */
#define NMID WM_USER + 8
/** @brief A debug message, execute this interrupt immediately */
#define RESETD WM_USER + 9
/** Waits until cpu stops if running and message PAUSECPU has been sent */
#define WAITEND WM_USER + 10
/** @brief Begins cpu execution */
#define STARTCPU WM_USER + 11
/** @brief Stops cpu execution */
#define PAUSECPU WM_USER + 12
/** @brief Sent by cpu plugin when cpu begins execution */
#define CPURUNNING WM_USER + 13
/** @brief Sent by cpu plugin when cpu stops */
#define CPUSTOPPED WM_USER + 14
/** @brief Turns off the video chip */
#define STOPVIDEO WM_USER + 15
#endif // USRMSGS_H